import org.w3c.dom.*;

//~--- JDK imports ------------------------------------------------------------

import java.io.*;

//Standard Java classes
import javax.swing.*;
import javax.swing.tree.*;

import javax.xml.xpath.*;

//JAXP
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

public class XTree extends JTree {
	// return null if 0 or > 1 nodes matching.
	public static Node findNode(String xpathStr, Document doc) {

		try {
			XPathFactory factory = XPathFactory.newInstance();
			XPath xpath = factory.newXPath();
			XPathExpression expr = null;

			expr = xpath.compile(xpathStr);
			Object result = expr.evaluate(doc, XPathConstants.NODESET);
			NodeList nodes = (NodeList) result;
			if (nodes.getLength() != 1) {
				System.out.println("0 or >1 node found\n");
				return null;
			}
			Node node = nodes.item(0);
			return node;
		} catch (javax.xml.xpath.XPathExpressionException xpe) {
			xpe.printStackTrace();
			return null;
		}
	}

	public static String generateXml(Node node) {
		try {
			Source source = new DOMSource(node);
			StringWriter stringWriter = new StringWriter();
			Result result = new StreamResult(stringWriter);
			TransformerFactory factory = TransformerFactory.newInstance();
			Transformer transformer = factory.newTransformer();
			transformer.transform(source, result);
			return stringWriter.getBuffer().toString();
		} catch (TransformerConfigurationException e) {
			e.printStackTrace();
		} catch (TransformerException e) {
			e.printStackTrace();
		}
		return null;
	}

	public static void updateNodeContent(Node node, String xpathStr,
			String newVal) {
		Jkmt jkmt = Jkmt.getInstance();
		String curVal = node.getTextContent();

		if (jkmt.net.protocol_state >= 5 && curVal != null && newVal != null && /*
												 * !"".equals(curVal) &&
												 * !"".equals(newVal) &&
												 */!curVal.equals(newVal)) {
			/* only enable USPs and print msg if its a real change */
			jkmt.consolePanels.printlnJkmt("Changing " + xpathStr + " from "
					+ curVal + " to " + newVal);
			UpdateServerPanel.enableAllUsp();
		}

		if (newVal != null)
			node.setTextContent(newVal); /* update DOM */

		jkmt.xmlEditorPanel.xTree.refresh(jkmt.xmlEditorPanel.xTree
				.generateXml()); /* refresh JTree to reflect DOM changes */

	}

	private DocumentBuilder db;

	private DocumentBuilderFactory dbf;

	public Document doc;

	/**
	 * This member stores the TreeNode object used to create the model for the
	 * JTree. The DefaultMutableTreeNode class is defined in the
	 * javax.swing.tree package and provides a default implementation of the
	 * MutableTreeNode interface.
	 */
	private DefaultMutableTreeNode treeNode;

	public XTree() throws ParserConfigurationException {

		// Initialize the superclass portion of the object
		super();

		// Set basic properties for the Tree rendering
		getSelectionModel().setSelectionMode(
				TreeSelectionModel.SINGLE_TREE_SELECTION);
		setShowsRootHandles(true);
		setEditable(false); // A more advanced version of this tool would allow
		// the Tree to be editable

		// Begin by initializing the object's DOM parsing objects
		dbf = DocumentBuilderFactory.newInstance();
		dbf.setValidating(false);
		db = dbf.newDocumentBuilder();

		// Take the DOM root node and convert it to a Tree model for the JTree
		setModel(buildWelcomeTree());
	} // end XTree()

	/**
	 * This constructor builds an XTree object using the XML text passed in
	 * through the constructor.
	 * 
	 * @param text
	 *            A String of XML formatted text
	 * 
	 * @exception ParserConfigurationException
	 *                This exception is potentially thrown if the constructor
	 *                configures the parser improperly. It won't.
	 */
	public XTree(String text) throws ParserConfigurationException {
		this();
		refresh(text);
	} // end XTree( String text )

	/**
	 * This method takes the XML text string passed to it, parses the XML and
	 * constructs the a DefaultTreeModel that can then be used to construct the
	 * graphical tree structure from that data
	 */
	private DefaultTreeModel buildTree(String text) {
		DefaultMutableTreeNode treeNode;
		Node newNode;

		// Take the DOM root node and convert it to a Tree model for the JTree
		newNode = parseXml(text);

		if (newNode != null) {
			treeNode = createTreeNode(newNode);

			return new DefaultTreeModel(treeNode);
		} else {
			return null;
		}
	} // end buildTree()

	/**
	 * This method builds a default DefaultTreeModel that can then be used to
	 * construct the graphical tree structure from that data
	 */
	private DefaultTreeModel buildWelcomeTree() {
		DefaultMutableTreeNode root, n;

		root = new DefaultMutableTreeNode("kmt-conf.xml editor");
		n = new DefaultMutableTreeNode("Not connected to server");
		root.add(n);

		return new DefaultTreeModel(root);
	}

	/**
	 * This takes a DOM Node and recurses through the children until each one is
	 * added to a DefaultMutableTreeNode. The JTree then uses this object as a
	 * tree model.
	 * 
	 * @param root
	 *            org.w3c.Node.Node
	 * 
	 * @return Returns a DefaultMutableTreeNode object based on the root Node
	 *         passed in
	 */
	private DefaultMutableTreeNode createTreeNode(Node root) {
		DefaultMutableTreeNode treeNode = null;
		String type, name, value;
		NamedNodeMap attribs;
		Node attribNode;

		// Get data from root node
		type = getNodeType(root);
		name = root.getNodeName();
		value = root.getNodeValue();

		// Special case for TEXT_NODE
		treeNode = new DefaultMutableTreeNode(
				(root.getNodeType() == Node.TEXT_NODE) ? value : name);

		// Display the attributes if there are any
		attribs = root.getAttributes();

		if (attribs != null) {
			for (int i = 0; i < attribs.getLength(); i++) {
				attribNode = attribs.item(i);
				name = attribNode.getNodeName().trim();
				value = attribNode.getNodeValue().trim();

				if (value != null) {
					if (value.length() > 0) {
						treeNode.add(new DefaultMutableTreeNode(
								"[Attribute] --> " + name + "=\"" + value
										+ "\""));
					} // end if ( value.length() > 0 )
				} // end if ( value != null )
			} // end for( int i = 0; i < attribs.getLength(); i++ )
		} // end if( attribs != null )

		// Recurse children nodes if any exist
		if (root.hasChildNodes()) {
			NodeList children;
			int numChildren;
			Node node;
			String data;

			children = root.getChildNodes();

			// Only recurse if Child Nodes are non-null
			if (children != null) {
				numChildren = children.getLength();

				for (int i = 0; i < numChildren; i++) {
					node = children.item(i);

					if (node != null) {

						// A special case could be made for each Node type.
						if (node.getNodeType() == Node.ELEMENT_NODE) {
							treeNode.add(createTreeNode(node));
						} // end if( node.getNodeType() == Node.ELEMENT_NODE )

						data = node.getNodeValue();

						if (data != null) {
							data = data.trim();

							if (!data.equals("\n") && !data.equals("\r\n")
									&& (data.length() > 0)) {
								treeNode.add(createTreeNode(node));
							} // end if ( !data.equals("\n") &&
							// !data.equals("\r\n") && data.length() > 0 )
						} // end if( data != null )
					} // end if( node != null )
				} // end for (int i=0; i < numChildren; i++)
			} // end if( children != null )
		} // end if( root.hasChildNodes() )

		return treeNode;
	} // end createTreeNode( Node root )

	public String generateXml() {
		return XTree.generateXml(this.doc);
	}

	/**
	 * This method returns a string representing the type of node passed in.
	 * 
	 * @param node
	 *            org.w3c.Node.Node
	 * 
	 * @return Returns a String representing the node type
	 */
	private String getNodeType(Node node) {
		String type;

		switch (node.getNodeType()) {
		case Node.ELEMENT_NODE: {
			type = "Element";

			break;
		}

		case Node.ATTRIBUTE_NODE: {
			type = "Attribute";

			break;
		}

		case Node.TEXT_NODE: {
			type = "Text";

			break;
		}

		case Node.CDATA_SECTION_NODE: {
			type = "CData section";

			break;
		}

		case Node.ENTITY_REFERENCE_NODE: {
			type = "Entity reference";

			break;
		}

		case Node.ENTITY_NODE: {
			type = "Entity";

			break;
		}

		case Node.PROCESSING_INSTRUCTION_NODE: {
			type = "Processing instruction";

			break;
		}

		case Node.COMMENT_NODE: {
			type = "Comment";

			break;
		}

		case Node.DOCUMENT_NODE: {
			type = "Document";

			break;
		}

		case Node.DOCUMENT_TYPE_NODE: {
			type = "Document type";

			break;
		}

		case Node.DOCUMENT_FRAGMENT_NODE: {
			type = "Document fragment";

			break;
		}

		case Node.NOTATION_NODE: {
			type = "Notation";

			break;
		}

		default: {
			type = "???";

			break;
		}
		} // end switch( node.getNodeType() )

		return type;
	} // end getNodeType()

	// NOTE: w3c Document (DOM) subclass of Node
	private Node parseXml(String xml) {
		ByteArrayInputStream byteStream;

		byteStream = new ByteArrayInputStream(xml.getBytes());

		try {
			doc = db.parse(byteStream);
		} catch (Exception e) {
			System.out.println(e.getMessage());
			return null;
		}

		return doc.getDocumentElement();
	} // end parseXml

	// update the JTree to this text
	public void refresh(String text) {
		TreeModel treeMod;

		// Take the DOM root node and convert it to a Tree model for the JTree
		treeMod = buildTree(text);

		if (treeMod != null) {
			setModel(treeMod);
		}

		for (int i = 0; i < this.getRowCount(); i++) {
			this.expandRow(i);
		}

	}

}
