/*
 *  DPS (Detect phony sys_call_table) dps.c 
 *  This code is specific to Linux only
 *  Primary Author: Ryan@innosecc.com
 *  Secondary Author: Sean@innosecc.com
 */

/* Summary:
 * This tool detects traditional Linux /dev/kmem rootkits
 * that modify the system call handler (which is invoked
 * through IDT 0x80).
 */


#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "../common/common.h"

/* We use packed attribute so that GCC */
/* does not do any zero padding for word */
/* alignment */

struct
{
  unsigned short limit;
  unsigned int base;
} __attribute__ ((packed)) idtr;

struct
{
  unsigned short off1;
  unsigned short sel;
  unsigned char none, flags;
  unsigned short off2;
} __attribute__ ((packed)) idt;

int main (int argc, char **argv)
{

	FILE *fd;
  	unsigned long sct_off;
  	unsigned long sct, sysmap_sct = 0;
  	unsigned char *p, code[255], buf[255];
  	int i, kmt_mem = 0;
	char kmem_path[255];
	
	if (argc < 2)
	{
		printf("DPS (Detect phony sys_call_table) (C) innosecc.com 2009\n");
		printf("Usage: %s <System.map> -M (For /dev/kmt_mem)\n", argv[0]);
		exit(-1);
	}
	
	if (argc > 2)
	{
		if(strcmp(argv[2], "-M") == 0)
			kmt_mem++;
	}
		
	if ((fd = fopen(argv[1], "r")) == NULL)
	{
		printf("[TOOL_FAILURE] Could not open %s - %s\n",argv[1], strerror(errno));
		exit(-1);
	}


	while(fgets(buf, sizeof(buf)-1, fd))
	{
		if (strstr(buf, "sys_call_table"))
		{
			p = strtok(buf, "rRtT");
			if (p)
			{
				sysmap_sct = strtoul(buf, NULL, 16);
				break;
			}
		}
	}

	if (!sysmap_sct)
	{
		printf("Fatal error: Unable to locate the sys_call_table in System.map\n");
		exit(-1);
	}

	asm ("sidt %0":"=m" (idtr));
		
	if (kmt_mem)
		strcpy(kmem_path, "/dev/kmt_mem");
	else
		strcpy(kmem_path, "/dev/kmem");

	if ((openkmem(kmem_path, O_RDONLY)) == -1)
	{
		printf("[TOOL_FAILURE] %s open %s\n", kmem_path, strerror(errno));
		exit(-1);
	}
	
	/* get int $0x80 */
	if (readkmem((unsigned char *)&idt, idtr.base + 8 * 0x80, sizeof (idt)) == -1)
	{
		printf("[TOOL_FAILURE] Failed to read from %s\n", kmem_path);
		exit(-1);
	}

	/* get system_call() */
	sct_off = (idt.off2 << 16) | idt.off1;
	
	if (readkmem(code, sct_off, 0x100) == -1)
	{
		printf("[TOOL_FAILURE] Failed to read from /dev/kmem\n");
		exit(-1);
	}

	/* find the code sequence that calls sys_call_table fp */
	sct = 0;
	for (i = 0; i < 255; i++)
	{
		if (code[i] == 0xff && code[i+1] == 0x14 && code[i+2] == 0x85)
			sct = code[i+3] + (code[i+4] << 8) + (code[i+5] << 16) + (code[i+6] << 24); 
	}
	
	closekmem();

	if (sct)
	{
		if (sct != sysmap_sct)
		{
			
			printf("[ALERT_TO_DAEMON]\n");
			printf("[!] Interrupt handler system_call() is using an evil sys_call_table: 0x%x\n", sct);
			printf("[!] The known good sys_call_table is 0x%x\n", sysmap_sct);
		}
		else
			//printf("The current sys_call_table being used is the correct one: 0x%x\n", sct);
			printf("[+] This Linux machine is using the correct sys_call_table: 0x%x\n", sct);
	}
	else
		printf("[!] Unable to locate sys_call_table symbol from within %s\n", argv[1]);
	
	exit(0);
}
