/*
 * kmt_mem (C) innosecc.com 2009
 * Author: Ryan O'Neill <ryan@innosecc.com>
 *
 * This code adds a device similar to /dev/kmem that allows for read access
 * from the .text segment in kernel memory, by KMT tools only.
 *
 * This is for Systems that do not have /dev/kmem read access compiled in
 * to the kernel.
 *
 * default major number is 250 unless module argument is given.
 * if kmtd_inode is ommitd security is disabled,otherwise only if the group_leader of the process trying to open kmt_mem's executable has that inode will it work. This is a seriously big pain to circumvent from userspace. The security of /dev/kmt_mem is 100% complete.
 *
 * insmod kmt_mem.ko kmt_major=111
 * insmod kmt_mem.ko kmt_major=111 kmtd_inode=`ls -i ./path/to/kmtd | awk '{ print $1; }'`
 * mknod /dev/kmt_mem c 250 0
 */

#include <linux/stat.h>
#include <linux/file.h>
#include <linux/uio.h>
#include <linux/smp_lock.h>
#include <linux/fsnotify.h>
#include <linux/security.h>
#include <linux/syscalls.h>
#include <linux/pagemap.h>
#include <linux/splice.h>

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>				/* printk() */
#include <linux/slab.h>					/* kmalloc() */
#include <linux/fs.h>						/* everything... */
#include <linux/errno.h>				/* error codes */
#include <linux/types.h>				/* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>				/* O_ACCMODE */


#include <linux/mm.h>						/* for VM_EXECUTABLE */
#include <linux/mount.h>				/* for mnt_get */
#include <linux/moduleparam.h>
#include <linux/sched.h>
/* for "current" ptr.
 * Please see
 * http://lkml.indiana.edu/hypermail/linux/kernel/0109.3/0991.html */

#include <asm/system.h>					/* cli(), *_flags */
#include <asm/uaccess.h>				/* copy_from/to_user */

#ifdef x86_64
#define _TEXT_START 0xc0000000	/* <- Magic address */
#define _TEXT_FINI  0xc0600000	/* <- Approx end of text */
#else
#define _TEXT_START 0xffffffff81000000
#define _TEXT_FINI  0xffffffff81800000
#endif


int mem_open (struct inode *inode, struct file *filp);
ssize_t mem_read (struct file *filp, char *buf, size_t count, loff_t * f_pos);
loff_t mem_lseek (struct file *filp, loff_t, int);

/* Global variables of the driver */
/* Major number */
int kmt_major = 250;
unsigned long kmtd_inode = 0;

unsigned char *mem;

module_param (kmt_major, int, 0);
module_param (kmtd_inode, long, 0);

struct file_operations kmt_fops = {
read:mem_read,
open:mem_open,
llseek:mem_lseek
		/* Make sure to NULL initialize unused ops */
		/*
		   readdir: NULL,
		   ioctl: NULL,
		   mmap: NULL,
		   fsync: NULL,
		   lock: NULL,
		   fasync: NULL,
		   flush: NULL,
		   release: NULL
		 */
};

/* A light implementation of llseek */
loff_t mem_lseek(struct file *file, loff_t offset, int origin)
{
        long long retval;
        struct inode *inode = file->f_mapping->host;

        mutex_lock(&inode->i_mutex);
        switch (origin) {
                case SEEK_END:
                        offset += inode->i_size;
                        break;
                case SEEK_CUR:
                        offset += file->f_pos;
        }
        retval = -EINVAL;
        if (offset>=0 && offset<=inode->i_sb->s_maxbytes) {
                if (offset != file->f_pos) {
                        file->f_pos = offset;
                        file->f_version = 0;
                }
                retval = offset;
        }
        mutex_unlock(&inode->i_mutex);
        return retval;
}

/* This is allows our user space apps to read() (i.e similar to read /dev/kmem) */
ssize_t
mem_read (struct file * filp, char *buf, size_t count, loff_t * f_pos)
{
	/* Make sure user does not want to read out of bounds of .text */

	/* Confirm that we don't read past text segment */
	if (count > _TEXT_FINI - *f_pos)
		return 0;
	/* Copy text data to user space */
	if (copy_to_user (buf, &mem[*f_pos - _TEXT_START], count))
		return -EFAULT;

	/* Increase offset */
	*f_pos += count;

	return count;

}

int
mem_open (struct inode *_inode, struct file *filp)
{
	// struct task_struct *task = NULL;
	pid_t pid = current->pid;
	pid_t pgid = current->group_leader->pid;
	char *comm = current->group_leader->comm;

	struct dentry *dentry = NULL;
	struct vfsmount *mnt = NULL;
	struct vm_area_struct *vma = NULL;
	unsigned long ino = 0;

	if (kmtd_inode == 0)
		return 0;										//success, not using optional security

	down_read (&current->group_leader->mm->mmap_sem);


	vma = current->group_leader->mm->mmap;
	while (vma)
	{
		if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file)
			break;
		vma = vma->vm_next;
	}

	if (vma)
	{
		mnt = mntget (vma->vm_file->f_path.mnt);
		dentry = dget (vma->vm_file->f_path.dentry);
	}

	up_read (&current->group_leader->mm->mmap_sem);

	if (dentry)
	{
		ino = dentry->d_inode->i_ino;
		/* inode is dentry->d_inode */
	}

	//generally comm same as argv[0], TODO: tes tif modifying argv[0] effects.

	printk ("kmt_mem: current->pid = %d, current->group_leader->pid = %d\n",
					pid, pgid);
	printk ("kmt_mem: group leader comm = %s, inode number = %ld\n", comm, ino);

/*
	if(strcmp(comm,"kmtd"))
	    return -EACCES;
*/

	if (0 != kmtd_inode && ino != kmtd_inode)
	{
		/* option kmtd_inode= was given and it doesnt match inode of group leader. TODO; make it look for inode of group leader */
		return -EACCES;
	}

	/* Success */
	return 0;
}

int
init_kmt_mem (void)
{

	int result;

	/* Registering device */
	if ((result = register_chrdev (kmt_major, "kmt_mem", &kmt_fops)) < 0)
	{
		printk ("kmt_mem: Unable to get major number %d\n", kmt_major);
		return result;
	}

	printk ("kmt_mem: Registered device kmt_mem major: %d\n", kmt_major);

	/* Assign pointer to text segment */
	mem = (unsigned char *) _TEXT_START;

	return 0;
}

void
close_kmt_mem (void)
{
	printk ("kmt_mem: Removing kmt_mem\n");
	unregister_chrdev (kmt_major, "kmt_mem");
}

module_init (init_kmt_mem);
module_exit (close_kmt_mem);
